#ifndef GPIO_H
#define GPIO_H

#include "common.h"

/** 
  * @brief General Purpose IO
  */

typedef struct {
  __IO uint32_t MODER;        /*!< GPIO port mode register,                                  Address offset: 0x00 */
  __IO uint16_t OTYPER;       /*!< GPIO port output type register,                           Address offset: 0x04 */
  uint16_t RESERVED0;         /*!< Reserved,                                                                 0x06 */
  __IO uint32_t OSPEEDR;      /*!< GPIO port output speed register,                          Address offset: 0x08 */
  __IO uint32_t PUPDR;        /*!< GPIO port pull-up/pull-down register,                     Address offset: 0x0C */
  __IO uint16_t IDR;          /*!< GPIO port input data register,                            Address offset: 0x10 */
  uint16_t RESERVED1;         /*!< Reserved,                                                                 0x12 */
  __IO uint16_t ODR;          /*!< GPIO port output data register,                           Address offset: 0x14 */
  uint16_t RESERVED2;         /*!< Reserved,                                                                 0x16 */
  __IO uint32_t BSRR;         /*!< GPIO port bit set/reset registerBSRR,                     Address offset: 0x18 */
  __IO uint32_t LCKR;         /*!< GPIO port configuration lock register,                    Address offset: 0x1C */
  __IO uint32_t AFR[2];       /*!< GPIO alternate function low register,                Address offset: 0x20-0x24 */
  __IO uint16_t BRR;          /*!< GPIO bit reset register,                                  Address offset: 0x28 */
  uint16_t RESERVED3;         /*!< Reserved,                                                                 0x2A */
} GPIO_Type;

#define TAHB2PERIPH_BASE       (0x48000000)
#define TGPIOA_BASE            (TAHB2PERIPH_BASE + 0x00000000)
#define TGPIOB_BASE            (TAHB2PERIPH_BASE + 0x00000400)
#define TGPIOC_BASE            (TAHB2PERIPH_BASE + 0x00000800)
#define TGPIOD_BASE            (TAHB2PERIPH_BASE + 0x00000C00)
#define TGPIOF_BASE            (TAHB2PERIPH_BASE + 0x00001400)

#define TGPIOA               ((GPIO_Type *) TGPIOA_BASE)
#define TGPIOB               ((GPIO_Type *) TGPIOB_BASE)
#define TGPIOC               ((GPIO_Type *) TGPIOC_BASE)
#define TGPIOD               ((GPIO_Type *) TGPIOD_BASE)
#define TGPIOF               ((GPIO_Type *) TGPIOF_BASE)

#define  TRCC_AHBENR_GPIOAEN   ((uint32_t)0x00020000)        /*!< GPIOA clock enable */
#define  TRCC_AHBENR_GPIOBEN   ((uint32_t)0x00040000)        /*!< GPIOB clock enable */
#define  TRCC_AHBENR_GPIOCEN   ((uint32_t)0x00080000)        /*!< GPIOC clock enable */
#define  TRCC_AHBENR_GPIODEN   ((uint32_t)0x00100000)        /*!< GPIOD clock enable */
#define  TRCC_AHBENR_GPIOFEN   ((uint32_t)0x00400000)        /*!< GPIOF clock enable */

typedef enum {
  GPIO_Mode_IN   = 0x00, /*!< GPIO Input Mode              */
  GPIO_Mode_OUT  = 0x01, /*!< GPIO Output Mode             */
  GPIO_Mode_AF   = 0x02, /*!< GPIO Alternate function Mode */
  GPIO_Mode_AN   = 0x03  /*!< GPIO Analog In/Out Mode      */
} GPIOMode_TypeDef;

typedef enum {
  GPIO_OType_PP = 0x00,
  GPIO_OType_OD = 0x01
} GPIOOType_TypeDef;

typedef enum {
  GPIO_Speed_Level_1  = 0x01, /*!< Medium Speed */
  GPIO_Speed_Level_2  = 0x02, /*!< Fast Speed   */
  GPIO_Speed_Level_3  = 0x03  /*!< High Speed   */
} GPIOSpeed_TypeDef;

typedef enum {
  GPIO_PuPd_NOPULL = 0x00,
  GPIO_PuPd_UP     = 0x01,
  GPIO_PuPd_DOWN   = 0x02
} GPIOPuPd_TypeDef;
/// Enum pro PortNumber
typedef enum {
  GpioPortA,
  GpioPortB,
  GpioPortC,
  GpioPortD,
  GpioPortF
} GpioPortNum;
/// Asociace port Adress a RCC clock
struct GpioAssocPort {
  GPIO_Type* portAdr;
  uint32_t   clkMask;
};
/** @file
  * @brief Obecný GPIO pin.
  * 
  * @class GpioClass
  * @brief Obecný GPIO pin.
  * 
  *  Ukázka přetížení operátorů. Návratové hodnoty jsou v tomto případě celkem zbytečné,
  * ale umožňují řetězení, takže je možné napsat např.
    @code
    +-+-+-led;
    @endcode
  * a máme na led 3 pulsy. Je to sice blbost, ale funguje.
  * Všechny metody jsou konstantní, protože nemění data uvnitř třídy.
  * Vlastně ani nemohou, protože data jsou konstantní.
*/
class GpioClass {
  public:
    /** Konstruktor
    @param port GpioPortA | GpioPortB | GpioPortC | GpioPortD | GpioPortF
    @param no   číslo pinu na portu
    @param type IN, OUT, AF, AN default OUT 
    */
    GpioClass (GpioPortNum const port, const uint32_t no, const GPIOMode_TypeDef type = GPIO_Mode_OUT);
    /// Nastav pin @param b na tuto hodnotu
    const GpioClass& operator<< (const bool b) const {
      if (b) io->BSRR = pos;
      else   io->BRR  = pos;
      return *this;
    }
//![Gpio example]
    /// Nastav pin na log. H
    const GpioClass& operator+ (void) const {
      io->BSRR = (uint32_t) pos;
      return *this;
    }
    /// Nastav pin na log. L
    const GpioClass& operator- (void) const {
      io->BRR  = pos;
      return *this;
    }
    /// Změň hodnotu pinu
    const GpioClass& operator~ (void) const {
      io->ODR ^= pos;
      return *this;
    };
    /// Načti logickou hodnotu na pinu
    const bool get (void) const {
      if (io->IDR & pos) return true;
      else               return false;
    };
    /// A to samé jako operátor
    const GpioClass& operator>> (bool& b) const {
      b = get();
      return *this;
    }
//![Gpio example]
    void setMode (GPIOMode_TypeDef p) {
      uint32_t dno = num * 2;
      io->MODER   &= ~(3UL << dno);
      io->MODER   |=  (p   << dno);
    }
    void setOType (GPIOOType_TypeDef p) {
      io->OTYPER  &= (uint16_t)~(1UL << num);
      io->OTYPER  |= (uint16_t) (p   << num);
    }
    void setSpeed (GPIOSpeed_TypeDef p) {
      uint32_t dno = num * 2;
      io->OSPEEDR &= ~(3UL << dno);
      io->OSPEEDR |=  (p   << dno);
    }
    void setPuPd (GPIOPuPd_TypeDef p) {
      uint32_t dno = num * 2;
      io->PUPDR   &= ~(3UL << dno);
      io->PUPDR   |=  (p   << dno);
    }
    void setAF (unsigned af) {
      register unsigned int pd,pn = num;
      pd = (pn & 7) << 2; pn >>= 3;
      io->AFR[pn] &= ~(0xFU << pd);
      io->AFR[pn] |=  (  af << pd);
    }
  private:
    /// Port.
    GPIO_Type* const io;
    /// A pozice pinu na něm, stačí 16.bit
    const uint16_t   pos;
    /// pro funkce setXXX necháme i číslo pinu
    const uint16_t   num;
  
};

#endif // GPIO_H
