/**
  ******************************************************************************
  * @file    system_stm32f0xx.c
  * @author  MCD Application Team
  * @version V1.0.0
  * @date    23-March-2012
  * @brief   CMSIS Cortex-M0 Device Peripheral Access Layer System Source File.
  *          This file contains the system clock configuration for STM32F0xx devices,
  *          and is customized for use with STM32F0-DISCOVERY Kit.
  *          The STM32F0xx is configured to run at 48 MHz, following the three
  *          configuration below:
  *            - PLL_SOURCE_HSI (default): HSI (~8MHz) used to clock the PLL, and
  *                                        the PLL is used as system clock source.
  *            - PLL_SOURCE_HSE          : HSE (8MHz) used to clock the PLL, and
  *                                        the PLL is used as system clock source.
  *            - PLL_SOURCE_HSE_BYPASS   : HSE bypassed with an external clock
  *                                        (8MHz, coming from ST-Link) used to clock
  *                                        the PLL, and the PLL is used as system
  *                                        clock source.
  *
  *
  * 1.  This file provides two functions and one global variable to be called from
  *     user application:
  *      - SystemInit(): Setups the system clock (System clock source, PLL Multiplier
  *                      and Divider factors, AHB/APBx prescalers and Flash settings),
  *                      depending on the configuration selected (see above).
  *                      This function is called at startup just after reset and
  *                      before branch to main program. This call is made inside
  *                      the "startup_stm32f0xx.s" file.
  *
  *      - SystemCoreClock variable: Contains the core clock (HCLK), it can be used
  *                                  by the user application to setup the SysTick
  *                                  timer or configure other parameters.
  *
  *      - SystemCoreClockUpdate(): Updates the variable SystemCoreClock and must
  *                                 be called whenever the core clock is changed
  *                                 during program execution.
  *
  * 2. After each device reset the HSI (8 MHz Range) is used as system clock source.
  *    Then SystemInit() function is called, in "startup_stm32f0xx.s" file, to
  *    configure the system clock before to branch to main program.
  *
  * 3. If the system clock source selected by user fails to startup, the SystemInit()
  *    function will do nothing and HSI still used as system clock source. User can
  *    add some code to deal with this issue inside the SetSysClock() function.
  *
  * 4. The default value of HSE crystal is set to 8MHz, refer to "HSE_VALUE" define
  *    in "stm32f0xx.h" file. When HSE is used as system clock source, directly or
  *    through PLL, and you are using different crystal you have to adapt the HSE
  *    value to your own configuration.
  *
  ******************************************************************************
  */


#include "rcc.h"
#include "flash.h"

/* Select the PLL clock source */

#define PLL_SOURCE_HSI        // HSI (~8MHz) used to clock the PLL, and the PLL is used as system clock source
//#define PLL_SOURCE_HSE        // HSE (8MHz) used to clock the PLL, and the PLL is used as system clock source
//#define PLL_SOURCE_HSE_BYPASS // HSE bypassed with an external clock (8MHz, coming from ST-Link) used to clock
// the PLL, and the PLL is used as system clock source

uint32_t SystemCoreClock    = 48000000;
const uint8_t AHBPrescTable[16] = {0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 6, 7, 8, 9};


static void SetSysClock (void);

/**
  * @brief  Setup the microcontroller system.
  *         Initialize the Embedded Flash Interface, the PLL and update the
  *         SystemCoreClock variable.
  * @param  None
  * @retval None
  */
void SystemInit (void) {
  /* Set HSION bit */
  RCC->CR.VAL    |= (uint32_t) 0x00000001;
  /* Reset SW[1:0], HPRE[3:0], PPRE[2:0], ADCPRE and MCOSEL[2:0] bits */
  RCC->CFGR.VAL  &= (uint32_t) 0xF8FFB80C;
  /* Reset HSEON, CSSON and PLLON bits */
  RCC->CR.VAL    &= (uint32_t) 0xFEF6FFFF;
  /* Reset HSEBYP bit */
  RCC->CR.VAL    &= (uint32_t) 0xFFFBFFFF;
  /* Reset PLLSRC, PLLXTPRE and PLLMUL[3:0] bits */
  RCC->CFGR.VAL  &= (uint32_t) 0xFFC0FFFF;
  /* Reset PREDIV1[3:0] bits */
  RCC->CFGR2.VAL &= (uint32_t) 0xFFFFFFF0;
  /* Reset USARTSW[1:0], I2CSW, CECSW and ADCSW bits */
  RCC->CFGR3.VAL &= (uint32_t) 0xFFFFFEAC;
  /* Reset HSI14 bit */
  RCC->CR2.VAL   &= (uint32_t) 0xFFFFFFFE;
  /* Disable all interrupts */
  RCC->CIR.VAL = 0x00000000;
  /* Configure the System clock frequency, AHB/APBx prescalers and Flash settings */
  SetSysClock();
}

/**
  * @brief  Update SystemCoreClock according to Clock Register Values
  *         The SystemCoreClock variable contains the core clock (HCLK), it can
  *         be used by the user application to setup the SysTick timer or configure
  *         other parameters.
  *
  * @note   Each time the core clock (HCLK) changes, this function must be called
  *         to update SystemCoreClock variable value. Otherwise, any configuration
  *         based on this variable will be incorrect.
  *
  * @note   - The system frequency computed by this function is not the real
  *           frequency in the chip. It is calculated based on the predefined
  *           constant and the selected clock source:
  *
  *           - If SYSCLK source is HSI, SystemCoreClock will contain the HSI_VALUE(*)
  *
  *           - If SYSCLK source is HSE, SystemCoreClock will contain the HSE_VALUE(**)
  *
  *           - If SYSCLK source is PLL, SystemCoreClock will contain the HSE_VALUE(**)
  *             or HSI_VALUE(*) multiplied/divided by the PLL factors.
  *
  *         (*) HSI_VALUE is a constant defined in stm32f0xx.h file (default value
  *             8 MHz) but the real value may vary depending on the variations
  *             in voltage and temperature.
  *
  *         (**) HSE_VALUE is a constant defined in stm32f0xx.h file (default value
  *              8 MHz), user has to ensure that HSE_VALUE is same as the real
  *              frequency of the crystal used. Otherwise, this function may
  *              have wrong result.
  *
  *         - The result of this function could be not correct when using fractional
  *           value for HSE crystal.
  * @param  None
  * @retval None
  */
void SystemCoreClockUpdate (void) {
  uint32_t prediv1factor, pllmull;

  switch (RCC->CFGR.BIT.SWS) {
    case USEHSI:  /* HSI used as system clock */
      SystemCoreClock = HSI_VALUE;
      break;
    case USEHSE:  /* HSE used as system clock */
      SystemCoreClock = HSE_VALUE;
      break;
    case USEPLL:  /* PLL used as system clock */
      /* Get PLL clock source and multiplication factor ----------------------*/
      pllmull = (uint32_t) RCC->CFGR.BIT.PLLMUL + 2;

      if (RCC->CFGR.BIT.PLLSRC == RESET) {
        /* HSI oscillator clock divided by 2 selected as PLL clock entry */
        SystemCoreClock = (HSI_VALUE >> 1) * pllmull;
      } else {
        prediv1factor = RCC->CFGR2.BIT.PREDIV + 1;
        /* HSE oscillator clock selected as PREDIV1 clock entry */
        SystemCoreClock = (HSE_VALUE / prediv1factor) * pllmull;
      }
      break;
    default: /* HSI used as system clock */
      SystemCoreClock = HSI_VALUE;
      break;
  }
  /* Compute HCLK clock frequency ----------------*/
  /* Get HCLK prescaler */
  pllmull = AHBPrescTable[RCC->CFGR.BIT.HPRE];
  /* HCLK clock frequency */
  SystemCoreClock >>= pllmull;
}

/**
  * @brief  Configures the System clock frequency, AHB/APBx prescalers and Flash
  *         settings.
  * @note   This function should be called only once the RCC clock configuration
  *         is reset to the default reset state (done in SystemInit() function).
  * @param  None
  * @retval None
  */
FLASH_Type * const FLASH = (FLASH_Type*) 0x40022000;

static void SetSysClock (void) {

  /* SYSCLK, HCLK, PCLK configuration ----------------------------------------*/
#if defined (PLL_SOURCE_HSI)
  /* At this stage the HSI is already enabled */

  /* Enable Prefetch Buffer and set Flash Latency */
  FLASH->ACR = FLASH_ACR_PRFTBE | FLASH_ACR_LATENCY;
  
  register RCC_CFGR_REG cfgr;
  cfgr.VAL = RCC->CFGR.VAL;
  cfgr.BIT.HPRE     = 0;
  cfgr.BIT.PPRE     = 0;
  cfgr.BIT.PLLSRC   = RESET;
  cfgr.BIT.PLLXTPRE = RESET;
  cfgr.BIT.PLLMUL   = RCC_CFGR_PLLMUL12;
  RCC->CFGR.VAL     = cfgr.VAL;
  /* Enable PLL */
  RCC->CR.BIT.PLLON = SET;

  /* Wait till PLL is ready */
  while ((RCC->CR.BIT.PLLRDY) == RESET);

  /* Select PLL as system clock source */
  RCC->CFGR.BIT.SW = USEPLL;

  /* Wait till PLL is used as system clock source */
  while (RCC->CFGR.BIT.SWS != USEPLL);
#else
#if defined (PLL_SOURCE_HSE)
  /* Enable HSE */
  RCC->CR.BIT.HSEON = SET;
#elif defined (PLL_SOURCE_HSE_BYPASS)
  /* HSE oscillator bypassed with external clock */
  RCC->CR.BIT.HSEON  = SET;
  RCC->CR.BIT.HSEBYP = SET;
#endif /* PLL_SOURCE_HSE */

  __IO uint32_t StartUpCounter = 0;
  __IO ONEBIT   HSEStatus;
  /* Wait till HSE is ready and if Time out is reached exit */
  do {
    HSEStatus = RCC->CR.BIT.HSERDY;
    StartUpCounter++;
  } while ((HSEStatus == RESET) && (StartUpCounter != HSE_STARTUP_TIMEOUT));

  HSEStatus = RCC->CR.BIT.HSERDY;

  if (HSEStatus == SET) {
    /* Enable Prefetch Buffer and set Flash Latency */
    FLASH->ACR = FLASH_ACR_PRFTBE | FLASH_ACR_LATENCY;

    register RCC_CFGR_REG cfgr;
    cfgr.VAL = RCC->CFGR.VAL;
    cfgr.BIT.HPRE     = 0;
    cfgr.BIT.PPRE     = 0;
    cfgr.BIT.PLLSRC   = SET;
    cfgr.BIT.PLLXTPRE = RESET;
    cfgr.BIT.PLLMUL   = RCC_CFGR_PLLMUL6;
    RCC->CFGR.VAL     = cfgr.VAL;
    /* Enable PLL */
    RCC->CR.BIT.PLLON = SET;

    /* Wait till PLL is ready */
    while ((RCC->CR.BIT.PLLRDY) == RESET);

    /* Select PLL as system clock source */
    RCC->CFGR.BIT.SW = USEPLL;

    /* Wait till PLL is used as system clock source */
    while (RCC->CFGR.BIT.SWS != USEPLL);
  } else {
    /* If HSE fails to start-up, the application will have wrong clock
         configuration. User can add here some code to deal with this error */
  }
#endif /* PLL_SOURCE_HSI */
}

